#include "Parser.h"

bool Parser::contains_help(std::string const &argument) {
    return argument == "--help" || argument == "-h";
}

void Parser::print_help() {
    std::cout << "ICP SLAM 1.0 is a SLAM software written as a part Lukas Majer Bachelor Thesis." << std::endl
              << "It loads dataset specified in [path] and performs ICP SLAM." << std::endl
              << "Dataset is a series of PNG images and corresponding depthmaps in binary format named" << std::endl
              << "'0..N.png' and '0..N.bin'. If odometry input is desired, dataset also has to contain timestamps"
              << std::endl
              << " in 'timestamps.csv' corresponding to each '*.png' and '*.bin' file " << std::endl
              << "and file with timestamped odometry readings odometry.txt." << std::endl
              << "written by Lukas Majer <majerlu2@fel.cvut.cz> " << std::endl;
    std::cout << "Usage: icp_slam [path] [ OPTIONS ]" << std::endl
              << "--help -h               Output this usage screen and exits" << std::endl
              << "--odometry -o        Uses additional odometry input" << std::endl;
}

bool Parser::process_arguments(int argc, char **argv) {
    bool should_terminate = false;
    if(argc < 2) {
        print_error();
        return true;
    }
    for (int i = 1; i < argc; i++) {
        if (contains_help(argv[i])) {
            print_help();
            return true;
        }
        if (i == 1) this->init.path = argv[1];
        if (contains_use_odmetry(argv[i])) {
            this->init.use_odometry = true;
        }
    }
    return should_terminate;
}

bool Parser::contains_use_odmetry(std::string const &argument) {
    return argument == "--odometry" || argument == "-o";
}

void Parser::print_error() {
    std::cerr << "Wrong input!" << std::endl;
}

