#include "Slam.h"

Slam::Slam() {
    icp.options.maxIterations = 100;
    icp.options.thresholdDist = 0.05;
    icp.options.thresholdAng = 0.01;
    icp.options.ALFA = 0.5;
    icp.options.smallestThresholdDist = 0.1;
    icp.options.onlyUniqueRobust = true;
    currentScan.colorScheme.scheme = CColouredPointsMap::cmFromIntensityImage;
    previousScan.colorScheme.scheme = CColouredPointsMap::cmFromIntensityImage;
    RUN_TIME = 0.0;
}

Slam::Slam(const Slam& orig) {
}

Slam::~Slam() {
}

void Slam::performSlam( string path,bool use_odometry) {
    float depthMap[SCAN_X][SCAN_Y];
    load_timestamps(path +"timestamps.csv");
    std::vector<data_entry> odometry;

    if(use_odometry) load_odometry(path + "odometry.csv",odometry);


    for (int i = 0; i < timestamps.size(); i++) {
        cout << "Processing scan number " << i << endl;
        CObservation3DRangeScanPtr obs = CObservation3DRangeScan::Create();

        ostringstream dep;
        dep << path  <<  i<< ".bin";

        ostringstream image;
        image << path <<  i<< ".png";
        
        Slam::loadDepthData(dep.str(), depthMap);
        Slam::makeTheScan(obs, image.str(), depthMap);
        
        currentScan.insertObservationPtr(obs);
        CPose3D mean;
        if (i == 0) {
            
        }
        if (i > 0) {
            CPose3D guess;
            if(use_odometry){
                if(i > 1)
              guess = find_matching_odometry(timestamps.at(i-1),odometry)-global;
            }
            Slam::matchScans(currentScan, previousScan, mean,guess);
            if(use_odometry){
                mean.z() = 0;
                mean.setYawPitchRoll(mean.yaw(),0,0);
            }
        }
        global = global + mean;

        std::cout << global << std::endl;
        transformations.push_back(global);
        previousScan = currentScan;
        currentScan.clear();

    }
    save_output(path);
    cout << "Running time: " << RUN_TIME << " seconds" << endl;
}

void Slam::makeTheScan(CObservation3DRangeScanPtr obs, string image, float (&depth) [SCAN_X][SCAN_Y]) {
    obs->intensityImage.loadFromFile(image);
    obs->hasIntensityImage = true;
    obs->rangeImage_setSize(SCAN_X, SCAN_Y);
    obs->resizePoints3DVectors(SCAN_X*SCAN_Y);

    for (int i = 0; i < SCAN_X; i++) {
        for (int j = 0; j < SCAN_Y; j++) {
                obs->rangeImage(i, j) = depth[i][j];
        }
    }
    obs->hasRangeImage = true;
    obs->range_is_depth = true;
    obs->cameraParams.cx(358.426);
    obs->cameraParams.cy(207.047);
    obs->cameraParams.fx(349.996);
    obs->cameraParams.fy(349.996);
    obs->hasPoints3D = true;
    obs->project3DPointsFromDepthImage();

}

void Slam::matchScans(CColouredPointsMap &referenceScan, CColouredPointsMap &alignedScan, CPose3D &calculatedPose,CPose3D &seed) {
    float runningTime;
    CICP::TReturnInfo icp_info;
    CPose3DPDFPtr pdf = icp.Align3D(
            &alignedScan,
            &referenceScan,
            seed,
            &runningTime,
            &icp_info
            );
    calculatedPose = pdf->getMeanVal();
    RUN_TIME+=(double) runningTime;
}

void Slam::loadDepthData(string depthName, float (&depth) [SCAN_X][SCAN_Y]) {
    FILE *myFile = fopen(depthName.c_str(), "r");
    float * buffer;
    buffer = (float *) malloc(sizeof (float)*SCAN_X * SCAN_Y);

    if (myFile != NULL) {
        fread(buffer, sizeof (float), SCAN_X * SCAN_Y, myFile);
        for (int i = 0; i < SCAN_X; i++) {
            for (int j = 0; j < SCAN_Y; j++) {
                    depth[i][j] = buffer[i * SCAN_Y + j] ;
            }
        }
        fclose(myFile);
    }
    free(buffer);
}

void Slam::load_timestamps(string input){
    ifstream fin(input);
    string line;
    if(fin.is_open()){
        while ( getline (fin,line) ) {
           timestamps.push_back(atol(line.c_str()));
            }
        fin.close();
    }
    
}
void Slam::save_output(string dest){
    ofstream outputFile(dest + "ICP_OUTPUT.csv");
  
    for (int i = 0; i < timestamps.size(); i++) {
        CQuaternion<double> quat;
        transformations.at(i).getAsQuaternion(quat);
        outputFile << timestamps.at(i)<< ", " << transformations.at(i).x() << ", "
                << transformations.at(i).y() << ", " << transformations.at(i).z() << ", " <<
                quat.x() << ", " << quat.y() << ", "<< quat.z() << ", "<< quat.r() << ", "<< endl;
    }       

    outputFile.close();

}

void Slam::load_odometry(string input,std::vector<data_entry> &dataset) {
    ifstream infile(input);
    double data[4];
    int index = 0;
    if (infile.is_open()) {
        string str;
        while (infile >> str){
            str.erase(std::remove(str.begin(), str.end(), ','), str.end());
            data[index]=atof(str.c_str());
            index++;
            if(index==4){
                CPose3D pose(data[1],data[2],0,data[3],0,0);
                size_t stamp = (size_t) data[0];
                data_entry temp;
                temp.timestamp = stamp;
                temp.data = pose;
                dataset.push_back(temp);
                index = index % 4;
                infile >> str;
                infile >> str;
            }
        }
    }
    CPose3D offset = dataset.front().data;
    for(auto it = dataset.begin();it != dataset.end();it++){
        CPose3D temp = it.operator*().data;
        temp = temp - CPose3D(0,0,0,DEG2RAD(90.0),0,0);
        it.operator*().data.x() = temp.x();
        it.operator*().data.y() = temp.y();
        it.operator*().data.z() = temp.z();
    }
}

CPose3D Slam::find_matching_odometry(size_t timestamp, std::vector<data_entry> &odometry) {
    for (auto it = odometry.begin();it != odometry.end();it++){
        if(it.operator*().timestamp >= timestamp) return it.operator*().data;
    }
    return CPose3D();
}

